"""
Examples for differential equations systems with dynamic control.

Here we provide several systems whose state is governed by differential
equations that allow us to plug in a controller. The controller receives
the state of the system as input and its output is added to one (or more) of
the differential equations.
"""
